#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020-2021, FICS, University of Florida
# All rights reserved.
# Jungmin Park
# Date : 10/27/2021
#
# ------------------------------------------------------------

import aes128_register as aes
import time

class AES128:
	""" AES128 object
	This class contain the public API for AES128 hardware implementation on ChipWhisperer FPGA board
	"""

	def __init__(self):
		import chipwhisperer as cw
		self.target = cw.target(None, cw.targets.CW305, fpga_id="100t")
		self.target.vccint_set(1.0)
		self.target.pll.pll_enable_set(True)
		self.target.pll.pll_outenable_set(False, 0)
		self.target.pll.pll_outenable_set(True, 1)
		self.target.pll.pll_outenable_set(False, 2)
		self.target.fpga_write(aes.REG_CLKSETTINGS, [0x1])

	def write_key(self, key):
		self.target.fpga_write(aes.REG_CRYPT_KEY, key)

	def write_textin(self, text):
		self.target.fpga_write(aes.REG_CRYPT_TEXTIN, text)

	def read_textout(self):
		data = self.target.fpga_read(aes.REG_CRYPT_CIPHEROUT, 16)
		return data

	def enc_128(self, key, plaintext):
		self.write_key(key)
		self.write_textin(plaintext)
		self.target.fpga_write(aes.REG_CRYPT_GO, [0x01])
		while self.target.fpga_read(aes.REG_CRYPT_GO, 1)[0] != 0x00:
			pass
		#time.sleep(1)
		data = self.read_textout()
		return data 	