#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020-2021, FICS, University of Florida
# Jungmin Park
# Date : 10/27/2021
# All rights reserved.
#
# ------------------------------------------------------------

import chipwhisperer as cw
import chipwhisperer.analyzer as cwa
from chipwhisperer.analyzer import aes_funcs
import time

proj = cw.open_project("aes128_EM2")

# Attack model 
# Hamming distance between round 9 and round 10
attack = cwa.cpa(proj, cwa.leakage_models.last_round_state_diff) 
attack.point_range = [2700,3000]

key = proj.keys[0]
correct_round_key = aes_funcs.key_schedule_rounds(key, 0, 10)

# This runs the attack:
start = time.time()
print("CPA Attacking ....")
attack_results = attack.run()
end = time.time()
print("Elapsed time: ", end='')
print("%d second" % (end-start))

print("10th round key : ")
for i in range(16):
    print("%02x " % correct_round_key[i], end='')
print('')

print("10th round guess key : ")
key_guess_10th = attack_results.key_guess()
success = 1
for i in range(16):
    print("%02x " % key_guess_10th[i], end='')
    if key_guess_10th[i] == correct_round_key[i]:
        print (" Correct!")
    else:
        print (" Wrong!")
        success = 0

print("Key guess : ")
key_guess = aes_funcs.key_schedule_rounds(key_guess_10th, 10, 0)
for i in range(16):
    print("%02x " % key_guess[i], end='')
print('')

if success == 1:
    print ("Successful CPA Attack!")
else:
    print ("CPA Attack Failed!")

